// SPDX-License-Identifier: GPL-2.0

#include "gcc-common.h"
#include "gcc-ksyms.h"

__visible int plugin_is_GPL_compatible;

static bool enabled;

static struct plugin_info arm_pertask_ssp_rtl_plugin_info = {
	.version	= "20210830",
	.help		= "arm_pertask_ssp_rtl plugin\n"
			  "enable\n"
};

static unsigned int arm_pertask_ssp_rtl_execute(void)
{
	rtx_insn *insn;

	for (insn = get_insns(); insn; insn = NEXT_INSN(insn)) {
		const char *sym;
		rtx body;
		rtx mask, masked_sp;

		/*
		 * Find a SET insn involving a SYMBOL_REF to __stack_chk_guard
		 */
		if (!INSN_P(insn))
			continue;
		body = PATTERN(insn);
		if (GET_CODE(body) != SET ||
		    GET_CODE(SET_SRC(body)) != SYMBOL_REF)
			continue;
		sym = XSTR(SET_SRC(body), 0);
		if (strcmp(sym, "__stack_chk_guard"))
			continue;

		/* ensure it's really the one we're looking for */
		gcc_assert(SYMBOL_REF_DECL(SET_SRC(body)) == default_stack_protect_guard());

		/* now we need these symbols for real! */
		gcc_assert(THREAD_STACK_MASK != NULL_TREE && TI_STACK_CANARY != NULL_TREE);

		/*
		 * Replace the source of the SET insn with an expression that
		 * produces the address of the copy of the stack canary value
		 * stored in struct thread_info
		 */
		mask = GEN_INT(sext_hwi(tree_to_shwi(THREAD_STACK_MASK), GET_MODE_PRECISION(Pmode)));
		masked_sp = gen_reg_rtx(Pmode);

		emit_insn_before(gen_rtx_set(masked_sp,
					     gen_rtx_AND(Pmode,
							 stack_pointer_rtx,
							 mask)),
				 insn);

		SET_SRC(body) = gen_rtx_PLUS(Pmode, masked_sp,
					     GEN_INT(tree_to_shwi(TI_STACK_CANARY)));

		df_insn_rescan(insn);
	}
	return 0;
}

static bool arm_pertask_ssp_rtl_gate(void)
{
	return flag_stack_protect != 0;
}

#define PASS_NAME arm_pertask_ssp_rtl
#if BUILDING_GCC_VERSION <= 4007
#define TODO_FLAGS_FINISH TODO_dump_func | TODO_df_verify
#else
#define TODO_FLAGS_FINISH TODO_dump_func | TODO_verify_rtl_sharing | TODO_df_verify
#endif
#include "gcc-generate-rtl-pass.h"

#if BUILDING_GCC_VERSION >= 9000
static void arm_pertask_ssp_start_unit(void *gcc_data, void *user_data)
{
	targetm.have_stack_protect_combined_set = hook_bool_void_false;
	targetm.have_stack_protect_combined_test = hook_bool_void_false;
}
#endif

__visible int plugin_init(struct plugin_name_args *plugin_info,
			  struct plugin_gcc_version *version)
{
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument *argv = plugin_info->argv;
	int i;

	PASS_INFO(arm_pertask_ssp_rtl, "expand", 1, PASS_POS_INSERT_AFTER);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	for (i = 0; i < argc; ++i) {
		if (!strcmp(argv[i].key, "enable")) {
			enabled = true;
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &arm_pertask_ssp_rtl_plugin_info);

	if (enabled) {
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &arm_pertask_ssp_rtl_pass_info);
#if BUILDING_GCC_VERSION >= 9000
		register_callback(plugin_name, PLUGIN_START_UNIT, arm_pertask_ssp_start_unit, NULL);
#endif
	}

	return 0;
}
