/*
 * Copyright 2019-2021 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Note: the choice of the license means that the compilation process is
 *       NOT 'eligible' as defined by gcc's library exception to the GPL v3,
 *       but for the kernel it doesn't matter since it doesn't link against
 *       any of the gcc libraries
 *
 * gcc plugin to convert k*alloc calls into type specific slab allocations
 *
 * TODO:
 * - too many to list
 *
 * BUGS:
 * - none known
 */

#include "autoslab.h"

__visible int plugin_is_GPL_compatible;

static struct plugin_info autoslab_plugin_info = {
	.version	= "202105190800",
	.help		= "verbose\tlog autoslab decisions\n"
			  "autostack\tenable autostack conversions\n"
};

bool verbose;
bool autostack;

static void find_autoslab_types(void *event_data __unused, void *data __unused)
{
	tree type = (tree)event_data;
	const_tree name;

	if (type == NULL_TREE || type == error_mark_node)
		return;

	if (TREE_CODE(type) != RECORD_TYPE )
		return;

	if (TYPE_FIELDS(type) == NULL_TREE)
		return;

	name = type_name(type);
	if (name == NULL_TREE)
		return;

	if (id_equal(name, "autoslab_info")) {
		if (autoslab_info_type) {
			gcc_assert(autoslab_info_type == type);
			return;
		}

		autoslab_info_type = type;
		build_pointer_type(type);
		build_pointer_type(TYPE_POINTER_TO(type));
	}

	if (id_equal(name, "autoslab_infos")) {
		if (autoslab_infos_type) {
			gcc_assert(autoslab_infos_type == type);
			return;
		}

		autoslab_infos_type = type;
		build_pointer_type(type);
	}
}

void autoslab_find_decls(void *event_data, void *data __unused)
{
	tree decl = (tree)event_data;
	enum kmem_call_names kmem_call;

	if (decl == error_mark_node)
		return;

	switch (TREE_CODE(decl)) {
	default:
		return;

	case FUNCTION_DECL:
		if (DECL_ARTIFICIAL(decl))
			return;

		if (DECL_ABSTRACT_ORIGIN(decl) && DECL_ABSTRACT_ORIGIN(decl) != decl)
			return;

		gcc_assert(DECL_NAME_POINTER(decl)[0]);

		if (autostack && id_equal(DECL_NAME(decl), CONVERT_AUTOSTACK)) {
			if (convert_autostack_decl == NULL_TREE) {
				tree argtype;

				argtype = TREE_TYPE(decl);
				gcc_assert(TREE_TYPE(argtype) == ptr_type_node);

				argtype = TYPE_ARG_TYPES(argtype);
				gcc_assert(TREE_VALUE(argtype) == ptr_type_node);
				argtype = TREE_CHAIN(argtype);
				gcc_assert(TYPE_MAIN_VARIANT(TREE_VALUE(argtype)) == size_type_node);
				argtype = TREE_CHAIN(argtype);
				gcc_assert(TYPE_MAIN_VARIANT(TREE_VALUE(argtype)) == unsigned_type_node);
				argtype = TREE_CHAIN(argtype);
				gcc_assert(TREE_VALUE(argtype) == TREE_VALUE(void_list_node));

				convert_autostack_decl = decl;
			} else
				gcc_assert(convert_autostack_decl == decl);
			return;
		}

		if (autostack && id_equal(DECL_NAME(decl), FREE_AUTOSTACK)) {
			if (free_autostack_decl == NULL_TREE) {
				tree argtype;

				argtype = TREE_TYPE(decl);
				gcc_assert(TREE_TYPE(argtype) == void_type_node);

				argtype = TYPE_ARG_TYPES(argtype);
				gcc_assert(TREE_VALUE(argtype) == const_ptr_type_node);
				argtype = TREE_CHAIN(argtype);
				gcc_assert(TYPE_MAIN_VARIANT(TREE_VALUE(argtype)) == size_type_node);
				argtype = TREE_CHAIN(argtype);
				gcc_assert(TREE_VALUE(argtype) == TREE_VALUE(void_list_node));

				free_autostack_decl = decl;
			} else
				gcc_assert(free_autostack_decl == decl);
			return;
		}

		if (id_equal(DECL_NAME(decl), "kfree")) {
			if (kfree_decl == NULL_TREE)
				kfree_decl = decl;
			else
				gcc_assert(kfree_decl == decl);
			return;
		}

		if (id_equal(DECL_NAME(decl), "kzfree")) {
			if (kzfree_decl == NULL_TREE)
				kzfree_decl = decl;
			else
				gcc_assert(kzfree_decl == decl);
			return;
		}

		if (id_equal(DECL_NAME(decl), "kfree_sensitive")) {
			if (kfree_sensitive_decl == NULL_TREE)
				kfree_sensitive_decl = decl;
			else
				gcc_assert(kfree_sensitive_decl == decl);
			return;
		}

		if (id_equal(DECL_NAME(decl), "kmem_cache_create_usercopy_typename")) {
			if (kmem_cache_create_usercopy_typename_decl == NULL_TREE)
				kmem_cache_create_usercopy_typename_decl = decl;
			else
				gcc_assert(kmem_cache_create_usercopy_typename_decl == decl);
			return;
		}

		for (kmem_call = KMEM_CACHE_ALLOC; kmem_call < KMEM_CACHE_NUM; ++kmem_call) {
			if (!id_equal(DECL_NAME(decl), kmem_cache_decls[kmem_call].name))
				continue;

			if (kmem_cache_decls[kmem_call].decl != NULL_TREE) {
				gcc_assert(decl == kmem_cache_decls[kmem_call].decl);
				continue;
			}

			kmem_cache_decls[kmem_call].decl = decl;
			return;
		}
	}
}

#if BUILDING_GCC_VERSION <= 4006
static void autoslab_find_all_decls(void *event_data __unused, void *data __unused)
{
	cgraph_node_ptr cnode;

	FOR_EACH_FUNCTION(cnode) {
		tree decl;

		if (NODE_THUNK(cnode) || cnode->alias)
			continue;
		if (NODE_GLOBAL(cnode)->inlined_to)
			continue;

		for (decl = NODE_DECL(cnode); decl && !convert_autostack_decl; decl = TREE_CHAIN(decl))
			autoslab_find_decls(decl, NULL);
		for (decl = NODE_DECL(cnode); decl && !free_autostack_decl; decl = TREE_CHAIN(decl))
			autoslab_find_decls(decl, NULL);
	}

	if (convert_autostack_decl)
		invoke_plugin_callbacks(PLUGIN_PRE_GENERICIZE, convert_autostack_decl);
	if (free_autostack_decl)
		invoke_plugin_callbacks(PLUGIN_PRE_GENERICIZE, free_autostack_decl);
}
#endif

static void register_attributes(void *event_data __unused, void *data __unused)
{
	register_attribute(&alloc_flags_attr);
	register_attribute(&alloc_node_attr);
	register_attribute(&alloc_attr);
}

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	int i;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;
	const char * const plugin_name = plugin_info->base_name;

	PASS_INFO(autotypename, "increase_alignment", 1, PASS_POS_INSERT_AFTER);
#if BUILDING_GCC_VERSION >= 4008
	PASS_INFO(autoslab, "slsr", 1, PASS_POS_INSERT_AFTER);
#elif BUILDING_GCC_VERSION >= 4006
	PASS_INFO(autoslab, "uninit", 1, PASS_POS_INSERT_AFTER);
#else
	PASS_INFO(autoslab, "*late_warn_uninitialized", 1, PASS_POS_INSERT_AFTER);
#endif

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	for (i = 0; i < argc; ++i) {
		if (!strcmp(argv[i].key, "verbose")) {
			verbose = true;
			continue;
		}

		if (!strcmp(argv[i].key, "autostack")) {
			autostack = true;
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &autoslab_plugin_info);
	register_callback(plugin_name, PLUGIN_FINISH_TYPE, find_autoslab_types, NULL);
#if BUILDING_GCC_VERSION <= 4006
	register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_START, autoslab_find_all_decls, NULL);
#else
	register_callback(plugin_name, PLUGIN_FINISH_DECL, autoslab_find_decls, NULL);
#endif
	register_callback(plugin_name, PLUGIN_PRE_GENERICIZE, autoslab_find_decls, NULL);
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &autotypename_pass_info);
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &autoslab_pass_info);
	register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_gt_autoslab);
	if (autostack)
		register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_gt_autostack);
	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);

	gcc_assert(!regcomp(&sizeof_regex, "^[[:space:]]*sizeof[[:space:]]*\\([[:space:]]*(struct|union)[[:space:]]*([_[:alnum:]]+)[[:space:]]*\\)[[:space:]]*$", REG_EXTENDED));
	return 0;
}
