/*
 * Copyright 2011 by Emese Revfy <re.emese@gmail.com>
 * Copyright 2011-2021 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * This gcc plugin implements a defense against memory corruption attacks by preventing unwanted
 * modifications of protected memory objects.
 *
 * This plugin is only concerned with objects with static storage duration, i.e., global variables whose
 * address is fixed at runtime (dynamically allocated objects are handled by KERNSEAL).
 *
 * Homepage:
 * http://www.grsecurity.net/~ephox/const_plugin/
 *
 * Usage:
 * $ gcc -I`gcc -print-file-name=plugin`/include -fPIC -shared -O2 -o constify_plugin.so constify_plugin.c
 * $ gcc -fplugin=constify_plugin.so test.c -O2
 *
 * TODO:
 * - add no_const to constified field if they're embedded in a non-constifiable type
 */

#include "constify.h"

__visible int plugin_is_GPL_compatible;

bool enabled = true;
bool verbose = false;

static struct plugin_info const_plugin_info = {
	.version	= "202102222222",
	.help		= "disable\tturn off constification\n",
};

static const struct {
	const char *name;
	const char *asm_op;
} const_sections[] = {
	{".text",			"\t.section\t.text,\"a\""},
	{".init.rodata",		"\t.section\t.init.rodata,\"a\""},
	{".ref.rodata",			"\t.section\t.ref.rodata,\"a\""},
	{".meminit.rodata",		"\t.section\t.meminit.rodata,\"a\""},
	{".memexit.rodata",		"\t.section\t.memexit.rodata,\"a\""},
	{RODATA_MUTABLE_READ_MOSTLY,	"\t.section\t" RODATA_MUTABLE_READ_MOSTLY ",\"a\""},
	{RODATA_MUTABLE,		"\t.section\t" RODATA_MUTABLE ",\"a\""},
	{".data.once",			"\t.section\t.data.once,\"a\""},
	{".pci_fixup_early",		"\t.section\t.pci_fixup_early,\"a\""},
	{".pci_fixup_header",		"\t.section\t.pci_fixup_header,\"a\""},
	{".pci_fixup_final",		"\t.section\t.pci_fixup_final,\"a\""},
	{".pci_fixup_enable",		"\t.section\t.pci_fixup_enable,\"a\""},
	{".pci_fixup_resume",		"\t.section\t.pci_fixup_resume,\"a\""},
	{".pci_fixup_resume_early",	"\t.section\t.pci_fixup_resume_early,\"a\""},
	{".pci_fixup_suspend",		"\t.section\t.pci_fixup_suspend,\"a\""},
	{".pci_fixup_suspend_late",	"\t.section\t.pci_fixup_suspend_late,\"a\""},
	{"__param",			"\t.section\t__param,\"a\""},
	{"__modver",			"\t.section\t__modver,\"a\""},
};

bool is_const_section(const char *name)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(const_sections); i++)
		if (!strcmp(const_sections[i].name, name))
			return true;

	return false;
}

/*
 * the TU must provide the open/close wrappers. look them up by name early enough and mark them as used
 * otherwise gcc would delete them if the TU itself doesn't use them and before the instrumentation needs them
 */
static void find_open_close_functions(void *event_data __unused, void *data __unused)
{
	tree fndecl = (tree)event_data;
	const char *asmname;

	if (fndecl == error_mark_node)
		return;

	if (TREE_CODE(fndecl) != FUNCTION_DECL)
		return;

	if (DECL_ARTIFICIAL(fndecl))
		return;

	if (DECL_ABSTRACT_ORIGIN(fndecl) && DECL_ABSTRACT_ORIGIN(fndecl) != fndecl)
		return;

	asmname = DECL_NAME_POINTER(fndecl);
	gcc_assert(asmname[0]);

	if (DECL_NAME_LENGTH(fndecl) == sizeof("pax_open_kernel") - 1 &&
	    id_equal(DECL_NAME(fndecl), "pax_open_kernel")) {
		gcc_assert(!pax_open_kernel_decl);
		pax_open_kernel_decl = fndecl;
		return;
	}

	if (DECL_NAME_LENGTH(fndecl) == sizeof("pax_close_kernel") - 1 &&
	    id_equal(DECL_NAME(fndecl), "pax_close_kernel")) {
		gcc_assert(!pax_close_kernel_decl);
		pax_close_kernel_decl = fndecl;
		return;
	}
}

static unsigned int (*old_section_type_flags)(tree decl, const char *name, int reloc);

static unsigned int constify_section_type_flags(tree decl, const char *name, int reloc)
{
	return is_const_section(name) ? 0 : old_section_type_flags(decl, name, reloc);
}

static bool (*old_lang_hooks_init)(void);

static bool constify_lang_hooks_init(void)
{
	register_callback("constify_plugin", PLUGIN_FINISH_TYPE, finish_type, NULL);
	return old_lang_hooks_init();
}

static tree (*old_merge_decl_attributes)(tree, tree);

static void constify_check_merged_decl_attributes(tree olddecl, tree newdecl)
{
	const char *oldname, *newname;
	bool oldattr, newattr;

	if (!VAR_P(olddecl) || !VAR_P(newdecl))
		return;

	oldname = get_decl_section_name(olddecl);
	newname = get_decl_section_name(newdecl);
	oldattr = !!lookup_attribute("mutable_const", DECL_ATTRIBUTES(olddecl));
	newattr = !!lookup_attribute("mutable_const", DECL_ATTRIBUTES(newdecl));

	if (oldattr == newattr) {
		if (!oldname || !newname)
			return;

		if (!strcmp(oldname, newname))
			return;
	}

	error_at(DECL_SOURCE_LOCATION(newdecl), "conflicting declarations for %qD [%s]", newdecl, newname ?: "<default>");
	inform(DECL_SOURCE_LOCATION(olddecl), "previous declaration of %q+D was here [%s]", olddecl, oldname ?: "<default>");
}

static tree constify_merge_decl_attributes(tree olddecl, tree newdecl)
{
	constify_check_merged_decl_attributes(olddecl, newdecl);
	return old_merge_decl_attributes(olddecl, newdecl);
}

static void constify_start_unit(void *gcc_data __unused, void *user_data __unused)
{
	old_section_type_flags = targetm.section_type_flags;
	targetm.section_type_flags = constify_section_type_flags;

	old_merge_decl_attributes = targetm.merge_decl_attributes;
	targetm.merge_decl_attributes = constify_merge_decl_attributes;
}

static void register_attributes(void *event_data __unused, void *data __unused)
{
	register_attribute(&no_const_attr);
	register_attribute(&do_const_attr);
	register_attribute(&mutable_const_attr);
}

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;
	int i;

	static const struct ggc_root_tab gt_ggc_r_gt_constify[] = {
		{
			.base = &pax_open_kernel_decl,
			.nelt = 1,
			.stride = sizeof(pax_open_kernel_decl),
			.cb = &gt_ggc_mx_tree_node,
			.pchw = &gt_pch_nx_tree_node
		},
		{
			.base = &pax_close_kernel_decl,
			.nelt = 1,
			.stride = sizeof(pax_close_kernel_decl),
			.cb = &gt_ggc_mx_tree_node,
			.pchw = &gt_pch_nx_tree_node
		},
		LAST_GGC_ROOT_TAB
	};

	PASS_INFO(check_local_variables, "ssa", 1, PASS_POS_INSERT_BEFORE);
	PASS_INFO(instrument_mutable_writes, "ssa", 1, PASS_POS_INSERT_AFTER);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	for (i = 0; i < argc; ++i) {
		if (!(strcmp(argv[i].key, "disable"))) {
			enabled = false;
			continue;
		}
		if (!(strcmp(argv[i].key, "verbose"))) {
			verbose = true;
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	if (strncmp(lang_hooks.name, "GNU C", 5) || !strncmp(lang_hooks.name, "GNU C+", 6)) {
		inform(UNKNOWN_LOCATION, G_("%s supports C only, not %s"), plugin_name, lang_hooks.name);
		enabled = false;
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &const_plugin_info);
	if (enabled) {
		old_lang_hooks_init = lang_hooks.init;
		lang_hooks.init = constify_lang_hooks_init;
		register_callback(plugin_name, PLUGIN_START_UNIT, constify_start_unit, (void *)plugin_name);
		register_callback(plugin_name, PLUGIN_PRE_GENERICIZE, find_open_close_functions, NULL);
		register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_START, handle_global_variables, NULL);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &check_local_variables_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &instrument_mutable_writes_pass_info);
		register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_gt_constify);
	}
	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);

	return 0;
}
