/*
 * Copyright 2011-2017 by Emese Revfy <re.emese@gmail.com>
 * Copyright 2019      by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2, or (at your option) v3
 *
 * Homepage:
 * https://github.com/ephox-gcc-plugins/sancov
 *
 * This plugin inserts a __sanitizer_cov_trace_pc() call at the start of basic blocks.
 * It supports all gcc versions with plugin support (from gcc-4.5 on).
 * It is based on the commit "Add fuzzing coverage support" by Dmitry Vyukov <dvyukov@google.com>.
 *
 * You can read about it more here:
 *  https://gcc.gnu.org/viewcvs/gcc?limit_changes=0&view=revision&revision=231296
 *  https://lwn.net/Articles/674854/
 *  https://github.com/google/syzkaller
 *  https://lwn.net/Articles/677764/
 *
 * Usage:
 * make run
 */

#include "gcc-common.h"

__visible int plugin_is_GPL_compatible;

static bool tracepc, tracecmp;

enum sancov_fns {
	SANCOV_TRACE_PC = 0,
	SANCOV_TRACE_CMP1,
	SANCOV_TRACE_CONST_CMP1,
	SANCOV_TRACE_CMP2,
	SANCOV_TRACE_CONST_CMP2,
	SANCOV_TRACE_CMP4,
	SANCOV_TRACE_CONST_CMP4,
	SANCOV_TRACE_CMP8,
	SANCOV_TRACE_CONST_CMP8,
	SANCOV_TRACE_CMPF,
	SANCOV_TRACE_CMPD,
	SANCOV_TRACE_SWITCH,
};

static struct {
	const char *name;
	tree decl;
} sancov_fns[] = {
	[SANCOV_TRACE_PC]         = { "__sanitizer_cov_trace_pc",         NULL_TREE },
	[SANCOV_TRACE_CMP1]       = { "__sanitizer_cov_trace_cmp1",       NULL_TREE },
	[SANCOV_TRACE_CONST_CMP1] = { "__sanitizer_cov_trace_const_cmp1", NULL_TREE },
	[SANCOV_TRACE_CMP2]       = { "__sanitizer_cov_trace_cmp2",       NULL_TREE },
	[SANCOV_TRACE_CONST_CMP2] = { "__sanitizer_cov_trace_const_cmp2", NULL_TREE },
	[SANCOV_TRACE_CMP4]       = { "__sanitizer_cov_trace_cmp4",       NULL_TREE },
	[SANCOV_TRACE_CONST_CMP4] = { "__sanitizer_cov_trace_const_cmp4", NULL_TREE },
	[SANCOV_TRACE_CMP8]       = { "__sanitizer_cov_trace_cmp8",       NULL_TREE },
	[SANCOV_TRACE_CONST_CMP8] = { "__sanitizer_cov_trace_const_cmp8", NULL_TREE },
	[SANCOV_TRACE_CMPF]       = { "__sanitizer_cov_trace_cmpf",       NULL_TREE },
	[SANCOV_TRACE_CMPD]       = { "__sanitizer_cov_trace_cmpd",       NULL_TREE },
	[SANCOV_TRACE_SWITCH]     = { "__sanitizer_cov_trace_switch",     NULL_TREE },
};

static struct plugin_info sancov_plugin_info = {
	.version	= "20190923",
	.help		= "sancov plugin\n"
			  "disable\n"
			  "trace-pc\n"
			  "trace-cmp\n"
};

static tree sancov_prepare_arg(gimple_stmt_iterator *gsi, tree arg, tree to_type, tree block, location_t loc)
{
	gimple_stmt_ptr cast_stmt;

	if (useless_type_conversion_p(to_type, TREE_TYPE(arg)))
		return arg;

	if (TREE_CODE(arg) == INTEGER_CST)
		return fold_convert_loc(loc, to_type, arg);

	cast_stmt = build_type_cast(to_type, arg, NULL_TREE);
	gimple_set_location(cast_stmt, loc);
	gimple_set_block(cast_stmt, block);
	gsi_insert_before(gsi, cast_stmt, GSI_SAME_STMT);
	return gimple_assign_lhs(cast_stmt);
}

static void sancov_cmp(gimple_stmt_iterator *gsi, tree lhs, tree rhs)
{
	tree type = TREE_TYPE(lhs);
	enum sancov_fns fncode;
	tree to_type, block;
	location_t loc;
	gimple_stmt_ptr call_stmt;

	if (INTEGRAL_TYPE_P(type)) {
		switch (int_size_in_bytes(type)) {
		case 1:
			fncode = SANCOV_TRACE_CMP1;
			to_type = unsigned_char_type_node;
			break;

		case 2:
			fncode = SANCOV_TRACE_CMP2;
			to_type = uint16_type_node;
			break;

		case 4:
			fncode = SANCOV_TRACE_CMP4;
			to_type = uint32_type_node;
			break;

		default:
			fncode = SANCOV_TRACE_CMP8;
			to_type = uint64_type_node;
			break;
		}

		if (is_gimple_min_invariant(rhs) != is_gimple_min_invariant(lhs)) {
			fncode = (enum sancov_fns)(fncode + 1);

			if (is_gimple_min_invariant(rhs)) {
				tree tmp = lhs;
				lhs = rhs;
				rhs = tmp;
			}
		}
	} else if (SCALAR_FLOAT_TYPE_P(type)) {
		if (TYPE_MODE(type) == TYPE_MODE(float_type_node)) {
			fncode = SANCOV_TRACE_CMPF;
			to_type = float_type_node;
		} else if (TYPE_MODE(type) == TYPE_MODE(double_type_node)) {
			fncode = SANCOV_TRACE_CMPD;
			to_type = double_type_node;
		} else
			return;
	} else
		return;

	loc = gimple_location(gsi_stmt(*gsi));
	block = gimple_block(gsi_stmt(*gsi));

	lhs = sancov_prepare_arg(gsi, lhs, to_type, block, loc);
	rhs = sancov_prepare_arg(gsi, rhs, to_type, block, loc);

	call_stmt = gimple_build_call(sancov_fns[fncode].decl, 2, lhs, rhs);
	gimple_set_location(call_stmt, loc);
	gimple_set_block(call_stmt, block);
	gsi_insert_before(gsi, call_stmt, GSI_SAME_STMT);
}

static void sancov_switch(gimple_stmt_iterator *gsi)
{
	gswitch *switch_stmt;
	HOST_WIDE_INT size_in_bytes;
	unsigned i, nlabels, ncases;
	location_t loc;
	tree block, case_array_type, case_array_var, ctor, index;
	gimple_stmt_ptr call_stmt;
	char case_array_name[32];
#if BUILDING_GCC_VERSION <= 4007
	VEC(constructor_elt, gc) *case_values;
#else
	vec<constructor_elt, va_gc> *case_values = NULL;
#endif
	static size_t case_array_count;

	switch_stmt = as_a_gswitch(gsi_stmt(*gsi));
	index = gimple_switch_index(switch_stmt);
	loc = gimple_location(switch_stmt);

	size_in_bytes = int_size_in_bytes(TREE_TYPE(index));
	if (size_in_bytes <= 0 || size_in_bytes > 8)
		return;

	nlabels = gimple_switch_num_labels(switch_stmt);
	ncases = 0;
	for (i = 1; i < nlabels; ++i) {
		tree label = gimple_switch_label(switch_stmt, i);

		if (CASE_LOW(label) != NULL_TREE)
			ncases++;
		if (CASE_HIGH(label) != NULL_TREE)
			ncases++;
	}

	if (!ncases)
		return;

#if BUILDING_GCC_VERSION <= 4007
	case_values = VEC_alloc(constructor_elt, gc, ncases + 2);
#else
	vec_alloc(case_values, ncases + 2);
#endif

	CONSTRUCTOR_APPEND_ELT(case_values, NULL_TREE, build_int_cst(uint64_type_node, ncases));
	CONSTRUCTOR_APPEND_ELT(case_values, NULL_TREE, build_int_cst(uint64_type_node, size_in_bytes * BITS_PER_UNIT));
	for (i = 1; i < nlabels; ++i) {
		tree case_value, label = gimple_switch_label(switch_stmt, i);

		case_value = CASE_LOW(label);
		if (case_value != NULL_TREE)
			CONSTRUCTOR_APPEND_ELT(case_values, NULL_TREE, fold_convert_loc(loc, uint64_type_node, case_value));

		case_value = CASE_HIGH(label);
		if (case_value != NULL_TREE)
			CONSTRUCTOR_APPEND_ELT(case_values, NULL_TREE, fold_convert_loc(loc, uint64_type_node, case_value));
	}

	case_array_type = build_array_type(build_type_variant(uint64_type_node, 1, 0), build_index_type(size_int(ncases + 2 - 1)));
	ctor = build_constructor(case_array_type, case_values);

	TREE_STATIC(ctor) = 1;
	TREE_READONLY(ctor) = 1;
	TREE_PUBLIC(ctor) = 0;
	TREE_CONSTANT(ctor) = 1;

	targetm.asm_out.generate_internal_label(case_array_name, "LCASEARRAY", case_array_count++);
	block = gimple_block(switch_stmt);
	case_array_var = build_decl(loc, VAR_DECL, get_identifier(case_array_name), case_array_type);

	DECL_INITIAL(case_array_var) = ctor;
	DECL_CONTEXT(case_array_var) = current_function_decl;
	DECL_ARTIFICIAL(case_array_var) = 1;
	DECL_EXTERNAL(case_array_var) = 0;
	DECL_IGNORED_P(case_array_var) = 1;

	TREE_STATIC(case_array_var) = 1;
	TREE_READONLY(case_array_var) = 1;
	TREE_PUBLIC(case_array_var) = 0;
	TREE_CONSTANT(case_array_var) = 1;
	TREE_ADDRESSABLE(case_array_var) = 1;
	TREE_USED(case_array_var) = 1;

	add_referenced_var(case_array_var);
	add_local_decl(cfun, case_array_var);
	varpool_add_new_variable(case_array_var);

	index = sancov_prepare_arg(gsi, index, uint64_type_node, block, loc);
	call_stmt = gimple_build_call(sancov_fns[SANCOV_TRACE_SWITCH].decl, 2, index, build_fold_addr_expr(case_array_var));
	gimple_set_location(call_stmt, loc);
	gimple_set_block(call_stmt, block);
	gsi_insert_before(gsi, call_stmt, GSI_SAME_STMT);
}

static unsigned int sancov_execute(void)
{
	basic_block bb;

	FOR_EACH_BB_FN(bb, cfun) {
		const_gimple_stmt_ptr stmt;
		gcall *gcall;
		gimple_stmt_iterator gsi = gsi_start_nondebug_after_labels_bb(bb);

		if (gsi_end_p(gsi))
			continue;

		if (tracepc) {
			stmt = gsi_stmt(gsi);
			gcall = as_a_gcall(gimple_build_call(sancov_fns[SANCOV_TRACE_PC].decl, 0));
			gimple_set_location(gcall, gimple_location(stmt));
			gimple_set_block(gcall, gimple_block(stmt));
			gsi_insert_before(&gsi, gcall, GSI_SAME_STMT);
		}

		if (!tracecmp)
			continue;

		for (; !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt = gsi_stmt(gsi);

			switch (gimple_code (stmt)) {
			case GIMPLE_ASSIGN: {
				enum tree_code rhs_code = gimple_assign_rhs_code(stmt);
				tree lhs, rhs;

				if (TREE_CODE_CLASS(rhs_code) == tcc_comparison) {
					lhs = gimple_assign_rhs1(stmt);
					rhs = gimple_assign_rhs2(stmt);
				} else if (rhs_code == COND_EXPR && COMPARISON_CLASS_P(gimple_assign_rhs1(stmt))) {
					tree cond = gimple_assign_rhs1(stmt);
					lhs = TREE_OPERAND(cond, 0);
					rhs = TREE_OPERAND(cond, 1);
				} else
					break;
				sancov_cmp(&gsi, lhs, rhs);
				break;
			}

			case GIMPLE_COND:
				sancov_cmp(&gsi, gimple_cond_lhs(stmt), gimple_cond_rhs(stmt));
				break;

			case GIMPLE_SWITCH:
				sancov_switch(&gsi);
				break;

			default:
				break;
			}

		}
	}
	return 0;
}

static bool sancov_gate(void)
{
	return !lookup_attribute("no_instrument_function", DECL_ATTRIBUTES(current_function_decl));
}

#define PASS_NAME sancov
#define TODO_FLAGS_FINISH TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_verify_flow
#include "gcc-generate-gimple-pass.h"

static void sancov_start_unit(void __unused *gcc_data, void __unused *user_data)
{
	unsigned int i;
	tree sancov_fn_types[] = {
		[SANCOV_TRACE_PC]         = build_function_type_list(void_type_node, NULL_TREE),
		[SANCOV_TRACE_CMP1]       = build_function_type_list(void_type_node, unsigned_char_type_node, unsigned_char_type_node, NULL_TREE),
		[SANCOV_TRACE_CONST_CMP1] = build_function_type_list(void_type_node, unsigned_char_type_node, unsigned_char_type_node, NULL_TREE),
		[SANCOV_TRACE_CMP2]       = build_function_type_list(void_type_node, uint16_type_node, uint16_type_node, NULL_TREE),
		[SANCOV_TRACE_CONST_CMP2] = build_function_type_list(void_type_node, uint16_type_node, uint16_type_node, NULL_TREE),
		[SANCOV_TRACE_CMP4]       = build_function_type_list(void_type_node, uint32_type_node, uint32_type_node, NULL_TREE),
		[SANCOV_TRACE_CONST_CMP4] = build_function_type_list(void_type_node, uint32_type_node, uint32_type_node, NULL_TREE),
		[SANCOV_TRACE_CMP8]       = build_function_type_list(void_type_node, uint64_type_node, uint64_type_node, NULL_TREE),
		[SANCOV_TRACE_CONST_CMP8] = build_function_type_list(void_type_node, uint64_type_node, uint64_type_node, NULL_TREE),
		[SANCOV_TRACE_CMPF]       = build_function_type_list(void_type_node, float_type_node,  float_type_node,  NULL_TREE),
		[SANCOV_TRACE_CMPD]       = build_function_type_list(void_type_node, double_type_node, double_type_node, NULL_TREE),
		[SANCOV_TRACE_SWITCH]     = build_function_type_list(void_type_node, uint64_type_node, ptr_type_node,    NULL_TREE),
	};
	tree nothrow_attr = tree_cons(get_identifier("nothrow"), NULL, NULL);
#if BUILDING_GCC_VERSION > 4005
	tree leaf_attr = tree_cons(get_identifier("leaf"), NULL, NULL);
#endif

	for (i = 0; i < ARRAY_SIZE(sancov_fns); i++) {
		tree sancov_fndecl;

		sancov_fndecl = build_fn_decl(sancov_fns[i].name, sancov_fn_types[i]);
		DECL_ASSEMBLER_NAME(sancov_fndecl);
		TREE_PUBLIC(sancov_fndecl) = 1;
		DECL_EXTERNAL(sancov_fndecl) = 1;
		DECL_ARTIFICIAL(sancov_fndecl) = 1;
		DECL_PRESERVE_P(sancov_fndecl) = 1;
		DECL_UNINLINABLE(sancov_fndecl) = 1;
		TREE_USED(sancov_fndecl) = 1;

		decl_attributes(&sancov_fndecl, nothrow_attr, 0);
		gcc_assert(TREE_NOTHROW(sancov_fndecl));
#if BUILDING_GCC_VERSION > 4005
		decl_attributes(&sancov_fndecl, leaf_attr, 0);
#endif

		sancov_fns[i].decl = sancov_fndecl;
		invoke_plugin_callbacks(PLUGIN_PRE_GENERICIZE, sancov_fndecl);
	}
}

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	int i;
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;

	static const struct ggc_root_tab gt_ggc_r_gt_sancov[] = {
		{
			.base = &sancov_fns[0].decl,
			.nelt = ARRAY_SIZE(sancov_fns),
			.stride = sizeof(sancov_fns[0]),
			.cb = &gt_ggc_mx_tree_node,
			.pchw = &gt_pch_nx_tree_node
		},
		LAST_GGC_ROOT_TAB
	};

	/* BBs can be split afterwards?? */
#if BUILDING_GCC_VERSION >= 4009
	PASS_INFO(sancov, "asan", 0, PASS_POS_INSERT_BEFORE);
#else
	PASS_INFO(sancov, "nrv", 1, PASS_POS_INSERT_BEFORE);
#endif

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	for (i = 0; i < argc; ++i) {
		if (!strcmp(argv[i].key, "disable")) {
			tracepc = tracecmp = false;
			continue;
		}
		if (!strcmp(argv[i].key, "trace-pc")) {
			tracepc = true;
			continue;
		}
		if (!strcmp(argv[i].key, "trace-cmp")) {
			tracecmp = true;
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &sancov_plugin_info);

	if (!tracepc && !tracecmp)
		return 0;

	register_callback(plugin_name, PLUGIN_START_UNIT, sancov_start_unit, NULL);
	register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_gt_sancov);
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &sancov_pass_info);

	return 0;
}
