/*
 * Copyright 2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * implement the 'no_sanitize' function attribute for gcc 4.9-7
 */

#include "utilities.h"

#if BUILDING_GCC_VERSION >= 4009 && BUILDING_GCC_VERSION < 8000

#if BUILDING_GCC_VERSION >= 4009 && BUILDING_GCC_VERSION < 6000
static const struct sanitizer_opts_s {
	const char * const name;
	unsigned int flag;
	size_t len;
	bool can_recover;
} sanitizer_opts[] = {
#define SANITIZER_OPT(name, flags, recover) { #name, flags, sizeof #name - 1, recover }
	SANITIZER_OPT(address, (SANITIZE_ADDRESS | SANITIZE_USER_ADDRESS), true),
	SANITIZER_OPT(thread, SANITIZE_THREAD, false),
	SANITIZER_OPT(undefined, SANITIZE_UNDEFINED, true),
	SANITIZER_OPT(all, ~0U, true),
#undef SANITIZER_OPT
	{ NULL, 0U, 0UL, false }
};
#endif

static unsigned int parse_no_sanitize_attribute(tree decl, char *value)
{
	unsigned int flags = 0;
	unsigned int i;
	char *q = strtok(value, ",");

	while (q != NULL) {
		for (i = 0; sanitizer_opts[i].name != NULL; ++i) {
			if (strcmp(sanitizer_opts[i].name, q))
				continue;

			flags |= sanitizer_opts[i].flag;
#if BUILDING_GCC_VERSION >= 5000
			if (sanitizer_opts[i].flag == SANITIZE_UNDEFINED)
				flags |= SANITIZE_NONDEFAULT;
#endif

			if ((flags & SANITIZE_ADDRESS) && !lookup_attribute("no_sanitize_address", DECL_ATTRIBUTES(decl)))
				DECL_ATTRIBUTES(decl) = tree_cons(get_identifier("no_sanitize_address"), NULL_TREE, DECL_ATTRIBUTES(decl));

			if ((flags & SANITIZE_UNDEFINED) && !lookup_attribute("no_sanitize_undefined", DECL_ATTRIBUTES(decl)))
				DECL_ATTRIBUTES(decl) = tree_cons(get_identifier("no_sanitize_undefined"), NULL_TREE, DECL_ATTRIBUTES(decl));

			if ((flags & SANITIZE_THREAD) && !lookup_attribute("no_sanitize_thread", DECL_ATTRIBUTES(decl)))
				DECL_ATTRIBUTES(decl) = tree_cons(get_identifier("no_sanitize_thread"), NULL_TREE, DECL_ATTRIBUTES(decl));
			break;
		}

		if (sanitizer_opts[i].name == NULL)
			warning_at(DECL_SOURCE_LOCATION(decl), OPT_Wattributes, "%<%s%> attribute directive ignored", q);

		q = strtok(NULL, ",");
	}

	return flags;
}

static void add_no_sanitize_value(tree decl, unsigned int flags)
{
	tree attr = lookup_attribute("no_sanitize", DECL_ATTRIBUTES(decl));

	if (attr) {
		unsigned int old_value = tree_to_uhwi(TREE_VALUE(attr));

		flags |= old_value;
		if (flags == old_value)
			return;

		TREE_VALUE(attr) = build_int_cst(unsigned_type_node, flags);
	} else
		DECL_ATTRIBUTES(decl) = tree_cons(get_identifier("no_sanitize"), build_int_cst(unsigned_type_node, flags), DECL_ATTRIBUTES(decl));
}

static tree handle_no_sanitize_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	unsigned int sanitize_flags = 0;
	tree decl;
	location_t loc;

	*no_add_attrs = true;
	decl = *node;
	loc = DECL_SOURCE_LOCATION(decl);

	if (TREE_CODE(decl) != FUNCTION_DECL) {
		warning_at(loc, OPT_Wattributes, "%qE attribute only applies to functions", name);
		return NULL_TREE;
	}

	for (; args; args = TREE_CHAIN(args)) {
		tree id;

		id = TREE_VALUE(args);
		if (TREE_CODE(id) != STRING_CST) {
			error_at(loc, "%qE argument not a string", name);
			return NULL_TREE;
		}

		sanitize_flags |= parse_no_sanitize_attribute(decl, ASTRDUP(TREE_STRING_POINTER(id)));
	}
	add_no_sanitize_value(decl, sanitize_flags);

	*no_add_attrs = false;
	return NULL_TREE;
}

struct attribute_spec no_sanitize_flags_attr= ATTRIBUTE_SPEC_INIT("no_sanitize", 1, -1, true, false, false, handle_no_sanitize_attribute, false, NULL);

// hooks
static typeof(flag_sanitize) flag_sanitize_orig;
static bool (*orig_option_can_inline_p)(tree, tree);

static bool no_sanitize_option_can_inline_p(tree caller, tree callee)
{
	flag_sanitize = flag_sanitize_orig;

	if (!orig_option_can_inline_p(caller, callee))
		return false;

	/* Allow inlining always_inline functions into no_sanitize_address functions.  */
	if (!lookup_attribute("no_sanitize_address", DECL_ATTRIBUTES(caller)))
		return true;
	if (lookup_attribute("no_sanitize_address", DECL_ATTRIBUTES(callee)))
		return true;
	if (!lookup_attribute("always_inline", DECL_ATTRIBUTES(callee)))
		return true;

	flag_sanitize &= ~SANITIZE_ADDRESS;
	return true;
}

void no_sanitize_register_hook(void)
{
	orig_option_can_inline_p = targetm.target_option.can_inline_p;
	targetm.target_option.can_inline_p = no_sanitize_option_can_inline_p;
}

static unsigned int no_sanitize_before_einline_execute(void)
{
	flag_sanitize_orig = flag_sanitize;
	return 0;
}

#define PASS_NAME no_sanitize_before_einline
#define NO_GATE
#define TODO_FLAGS_FINISH TODO_dump_func
#include "gcc-generate-gimple-pass.h"

static unsigned int no_sanitize_after_einline_execute(void)
{
	flag_sanitize = flag_sanitize_orig;
	return 0;
}

#define PASS_NAME no_sanitize_after_einline
#define NO_GATE
#define TODO_FLAGS_FINISH TODO_dump_func
#include "gcc-generate-gimple-pass.h"

static unsigned int no_sanitize_before_inline_execute(void)
{
	flag_sanitize_orig = flag_sanitize;
	return 0;
}

#define PASS_NAME no_sanitize_before_inline
#define NO_WRITE_SUMMARY
#define NO_GENERATE_SUMMARY
#define NO_READ_SUMMARY
#define NO_READ_OPTIMIZATION_SUMMARY
#define NO_WRITE_OPTIMIZATION_SUMMARY
#define NO_STMT_FIXUP
#define NO_FUNCTION_TRANSFORM
#define NO_VARIABLE_TRANSFORM
#define NO_GATE

#include "gcc-generate-ipa-pass.h"

static unsigned int no_sanitize_after_inline_execute(void)
{
	flag_sanitize = flag_sanitize_orig;
	return 0;
}

#define PASS_NAME no_sanitize_after_inline
#define NO_WRITE_SUMMARY
#define NO_GENERATE_SUMMARY
#define NO_READ_SUMMARY
#define NO_READ_OPTIMIZATION_SUMMARY
#define NO_WRITE_OPTIMIZATION_SUMMARY
#define NO_STMT_FIXUP
#define NO_FUNCTION_TRANSFORM
#define NO_VARIABLE_TRANSFORM
#define NO_GATE

#include "gcc-generate-ipa-pass.h"

#endif
